@@ -42,6 +42,16 @@ class AxesProxyHandler(AxesHandler):
             cls.implementation = import_string(settings.AXES_HANDLER)()
         return cls.implementation
 
+    @classmethod
+    def reset_attempts(cls, *, ip_address: str = None, username: str = None) -> int:
+        return cls.get_implementation().reset_attempts(
+            ip_address=ip_address, username=username
+        )
+
+    @classmethod
+    def reset_logs(cls, *, age_days: int = None) -> int:
+        return cls.get_implementation().reset_logs(age_days=age_days)
+
     @staticmethod
     def update_request(request):
         """
@@ -49,15 +59,18 @@ class AxesProxyHandler(AxesHandler):
         """
 
         if request is None:
-            log.error('AXES: AxesProxyHandler.update_request can not set request attributes to a None request')
+            log.error(
+                "AXES: AxesProxyHandler.update_request can not set request attributes to a None request"
+            )
             return
-
-        request.axes_locked_out = False
-        request.axes_attempt_time = now()
-        request.axes_ip_address = get_client_ip_address(request)
-        request.axes_user_agent = get_client_user_agent(request)
-        request.axes_path_info = get_client_path_info(request)
-        request.axes_http_accept = get_client_http_accept(request)
+        if not hasattr(request, "axes_updated"):
+            request.axes_locked_out = False
+            request.axes_attempt_time = now()
+            request.axes_ip_address = get_client_ip_address(request)
+            request.axes_user_agent = get_client_user_agent(request)
+            request.axes_path_info = get_client_path_info(request)
+            request.axes_http_accept = get_client_http_accept(request)
+            request.axes_updated = True
 
     @classmethod
     def is_locked(cls, request, credentials: dict = None) -> bool:
@@ -69,11 +82,18 @@ class AxesProxyHandler(AxesHandler):
         cls.update_request(request)
         return cls.get_implementation().is_allowed(request, credentials)
 
+    @classmethod
+    def get_failures(cls, request, credentials: dict = None) -> int:
+        cls.update_request(request)
+        return cls.get_implementation().get_failures(request, credentials)
+
     @classmethod
     @toggleable
     def user_login_failed(cls, sender, credentials: dict, request=None, **kwargs):
         cls.update_request(request)
-        return cls.get_implementation().user_login_failed(sender, credentials, request, **kwargs)
+        return cls.get_implementation().user_login_failed(
+            sender, credentials, request, **kwargs
+        )
 
     @classmethod
     @toggleable
